---------------------------------------------------------------------------------------------
--[[
	qOOPLib - Quick Object Oriented Programming Library
    Copyright (C) 2011.
	Written by Roland Yonaba - E-mail: roland[dot]yonaba[at]gmail[dot]com
	qOOPLib is a simple and quick-written Library which offers some OOP facilities using Lua.
	Version: 1.2 QuickFix
	Project Page : http://qooplua.codeplex.com/
	Current Features :
		- Class creation with default methods granted
		- Static class
		- Final class
		- Single Inheritance
		- Multiple Inheritance
		- Instantiation with class() call
		- Instantiation with class(t) call and named arguments
		- Access to super Methods

--]]
----------------------------------------------------------------------------------------------

-- Hold the list of reserved keywords
-- They are all part of any class, and should neither be propagated from a superclass to its subclasses
-- nor from a class to its instances
local reserved_keywords = {'__super',',__sub','__index','__abstract','__final','__class'}

-- Searches a method in a list of tables. This will be the core
-- of multiple inheritance feature implementation
local function call_method(method,...)
	for k in ipairs(arg) do
	local m = arg[k][method]
		if m then return m end
	end
end

-- Searches an item in a table.
-- Returns true on success,plus the key, otherwise false.
local function table_find(item,table)
	for k in pairs(table) do
		if table[k] == item then return true,k end
	end
	return false
end

-- Returns a reference-independant copy of a table
-- Functions and reserved_keywords entries are not copied.
local function table_copy(to,from)
	for k,v in pairs(from) do
		if not table_find(k,reserved_keywords) then
			if type(v) ~='function' then
				if type(v) == 'table' then
					if is_Object(v) then
					local super = getmetatable(v)
					to[k] = super()
					else
					to[k]={}
					table_copy(to[k],v)
					end
				else
					to[k] = v
				end
			end
		end
	end
end

-- Generates a new instance of a class
local function instance(class)
	local o = {}
	table_copy(o,class)
	return setmetatable(o,class)
end

-- Checks if 'class' is a class
local function is_Class(class)                            --ok
	local c = class.__class
	if c then
		return (c == class)
	else
		return false
	end
end

-- Checks if 'o' is an instance
local function is_Object(o)                               --ok
	local mt = getmetatable(o)
	if type(o)~='table' or not mt then return false end
	local check = false
		for i,reserved in ipairs(reserved_keywords) do
			check = table_find(reserved,reserved_keywords)
			if not check then return false end
		end
	if rawget(o,'getClass') or rawget(o,'getSuperClass') then return false end
	return check
end


-- Checks if 'o' is an instance of 'class'
local function is_ObjectOf(o,class)                               --ok
	if not is_Class(class) then error('class is not a qOOP class!',2) end
	if not o.getClass then error('object is not a qOOP object!',2) end
		local c = o:getClass()
		return c == class
end

-- Checks class.method existence
local function is_MemberOf(member,class)
	if not is_Class(class) then error('class is not a qOOP class!',2) end
		return (class[member] and true or false)
end

-- Inits 'class' members/extra-members
local function setMembers(class,arg)
	if not is_Class(class) then error('class is not a qOOP class!',2) end
	table_copy(class,arg)
end


-- Iterates over the subs of a class
	local function subs(t)
		if not is_Class(t) then error('class is not a qOOP class!',2) end
		local child = 0
		local num_child = #t.__sub
			return function()
				child = child + 1
				if child <= num_child then return t.__sub[child] end
				end
	end

-- Iterates over the superclasses of a class
	local function supers(t)
		if not is_Class(t) then error('class is not a qOOP class!',2) end
		local super = 0
		local num_super = #t.__super
			return function()
					super = super + 1
					if super <= num_super then return t.__super[super] end
					end
	end



-- Creates and returns 'class' with single, multiple inheritance
local function class(...)
	local new_class = {}
	local super = {}
		if arg and #arg > 0 then
			for i,parent in ipairs(arg) do
				if parent.__final then error('Arg #'..i..' is a final qOOP class!',2) end
			table.insert(super,parent)
			table.insert(parent.__sub,new_class)
			end
		end
	new_class.__super = super
	new_class.__sub = {}

	setmetatable(new_class,{__index = function (t,m)
							return call_method(m,unpack(new_class.__super))
							end,
							__call = function(self,t)
								if self.__abstract then error('Cannot instantiate from abstract qOOP class!',2) end
								local object = instance(self)
									if t then table_copy(object,t) end
							return object
							end
							})

	new_class.__index = new_class
	new_class.__abstract = false
	new_class.__final = false
	new_class.__class = new_class

	function new_class:setMembers(arg)                                --ok
		if not is_Class(self) then error('setMembers() must be called from a qOOP class, not an instance!',2) end
		setMembers(self,arg)
	end


	function new_class:hasMember(member)                                --ok
		return is_MemberOf(member,self)
	end

	function new_class:subclassOf(class)   --ok
		if not is_Class(self) then error('subclassOf() must be called from a qOOP class!',2) end
		return (table_find(class,self.__super))
	end

	function new_class:superOf(class)   --ok
		if not is_Class(self) then error('superOf() must be called from a qOOP class!',2) end
		return (table_find(class,self.__sub))
	end


	function new_class:getClass()                               --ok
		if not is_Object(self) then error('getClass() must be called from a qOOP Object!',2) end
		return new_class
	end

	function new_class:getSupers()
		if not is_Class(self) then error('getSuperclass() must be called from a qOOP class!',2) end
		return (new_class.__super)
	end

	function new_class:getSubs()   --ok
		if not is_Class(self) then error('getSubclass() must be called from a qOOP class!',2) end
		return (new_class.__sub)
	end

	function new_class:abstract()                               --ok
		if not is_Class(self) then error('abstract() must be called from a qOOP class!',2) end
		self.__abstract = true
	end

	function new_class:final()  --ok
		if not is_Class(self) then error('final() must be called from a qOOP class!',2) end
		self.__final = true
	end

	function new_class:super(method,...)   --ok
		local parent = self.__super
			if #parent > 0 then
			local m = call_method(method,unpack(parent))
				if m then
				m(self,unpack(arg))
				else
					error('Cannot find method in parents!',2)
				end
			else
				error('qOOP Class has no parent!',2)
			end
	end
	return new_class
end


return { 
			isClass = is_Class,
			isObject = is_Object,
			isObjectOf = is_ObjectOf,
			children = subs,
			parents = supers,
			class = class,
		}
